"""
Classes for wrapping HDF5 datasets.
"""
from typing import Union, AbstractSet, Optional, Tuple, Any, Dict

import numpy as np
import pandas as pd
import h5py as h5

from audata import _utils as utils
from audata.element import Element


class Dataset(Element):
    """
    Maps to an HDF5 dataset, maintaining the `audata` schema and facilitating translation of
    higher-level data types. Generally should not be instantiated directly.
    """

    def __init__(self, au_parent: Element, name: str):
        if not isinstance(au_parent, Element):
            raise Exception(f'Invalid parent: {type(au_parent)}')

        parent = au_parent.hdf
        if not isinstance(parent, h5.Group):
            raise Exception(f'Invalid parent: {type(parent)}')

        if name not in parent or not isinstance(parent[name], h5.Dataset):
            raise Exception(
                f'Path {name} is not a dataset in {parent.file.filename}:{parent.name}'
            )

        super().__init__(au_parent, name)

    @classmethod
    def new(cls,
            au_parent: Element,
            name: str,
            value: Union[h5.Dataset, np.ndarray, np.recarray, pd.DataFrame],
            overwrite: bool = False,
            **kwargs) -> 'Dataset':
        """Create a new Dataset object."""

        if not isinstance(au_parent, Element):
            raise Exception('Must send Element.')

        parent = au_parent.hdf
        if not isinstance(parent, h5.Group):
            raise Exception(f'Invalid parent: {type(parent)}')

        if name in parent and not overwrite:
            raise Exception(f'{name} already exists.')

        # If given an HDF5 dataset or an Dataset, read in all of its data
        # (HDF5 dataset as a numpy ndarray, Dataset as a pandas DataFrame).
        if isinstance(value, (h5.Dataset, Dataset)):
            value = value[:]

        # Try to create a class now.

        if isinstance(value, (np.ndarray, np.recarray)):
            return cls.__new_from_array(au_parent, name, value, **kwargs)

        elif isinstance(value, pd.DataFrame):
            return cls.__new_from_dataframe(au_parent, name, value)

        else:
            raise Exception(f'Unsure how to convert type {type(value)}')

    @classmethod
    def __new_from_array(cls,
                         au_parent: Element,
                         name: str,
                         arr: Union[np.ndarray, np.recarray],
                         time_cols: Optional[AbstractSet[str]] = None,
                         timedelta_cols: Optional[AbstractSet[str]] = None
                        ) -> 'Dataset':
        """Create a new dataset from a numpy recarray or ndarray."""

        if time_cols is None:
            time_cols = set({})
        if timedelta_cols is None:
            timedelta_cols = set({})

        # ATW: TODO: Less lame.
        if arr.dtype.names is None:
            return cls.__new_from_dataframe(au_parent, name,
                                            pd.DataFrame(data=arr))

        meta, recs = utils.audata_from_arr(
            arr,
            time_ref=au_parent.file.time_reference,
            time_cols=time_cols,
            timedelta_cols=timedelta_cols)
        au_parent.hdf.create_dataset(name,
                                     chunks=True,
                                     maxshape=(None,),
                                     compression='gzip',
                                     shuffle=True,
                                     fletcher32=True,
                                     data=recs)
        au_parent.hdf[name].attrs['.meta'] = utils.dict2json(meta)
        dataset = cls(au_parent, name)
        return dataset

    @classmethod
    def __new_from_dataframe(cls,
                             au_parent: Element,
                             name: str,
                             data: pd.DataFrame,
                             time_cols: Optional[AbstractSet[str]] = None,
                             timedelta_cols: Optional[AbstractSet[str]] = None
                            ) -> 'Dataset':
        """Create a new dataset from a pandas DataFrame."""

        if time_cols is None:
            time_cols = set({})
        if timedelta_cols is None:
            timedelta_cols = set({})

        meta, recs = utils.audata_from_df(
            data,
            time_ref=au_parent.file.time_reference,
            time_cols=time_cols,
            timedelta_cols=timedelta_cols)
        au_parent.hdf.create_dataset(name,
                                     chunks=True,
                                     maxshape=(None,),
                                     compression='gzip',
                                     shuffle=True,
                                     fletcher32=True,
                                     data=recs)
        au_parent.hdf[name].attrs['.meta'] = utils.dict2json(meta)
        dataset = cls(au_parent, name)
        return dataset

    def __getitem__(self, idx=slice(-1)) -> pd.DataFrame:
        return self.get(idx)

    def get(self,
            idx=slice(-1),
            raw: bool = False,
            datetimes: Optional[bool] = None) -> pd.DataFrame:
        """Return a dataset as a pandas DataFrame."""

        rec = self.hdf[idx]
        if raw:
            return rec
        if isinstance(rec, np.void):
            rec = np.array([rec], dtype=rec.dtype)

        if datetimes is None:
            datetimes = self.file.return_datetimes
        data = utils.df_from_audata(rec, self.columns, self.file.time_reference, datetimes)
        return data

    def append(self,
               data: Union[pd.DataFrame, np.recarray],
               direct: bool = False,
               time_cols: Optional[AbstractSet[str]] = None,
               timedelta_cols: Optional[AbstractSet[str]] = None):
        """Append additional data to a dataset."""

        if time_cols is None:
            time_cols = set({})
        if timedelta_cols is None:
            timedelta_cols = set({})

        arr = None
        if isinstance(data, np.recarray):
            arr = data
            if not direct:
                _, arr = utils.audata_from_arr(arr,
                                               time_ref=self.time_reference,
                                               time_cols=time_cols,
                                               timedelta_cols=timedelta_cols)
        elif direct:
            raise ValueError(
                ('Data must be in a recarray already to use direct append! '
                 f'Instead {type(data)} was sent.'))
        elif isinstance(data, pd.DataFrame):
            # Issue 3: This does not validate categorical levels. So keep them the same!
            _, arr = utils.audata_from_df(data,
                                          time_ref=self.time_reference,
                                          time_cols=time_cols,
                                          timedelta_cols=timedelta_cols)

        data_len = len(arr)
        self.hdf.resize((self.nrow + data_len,))
        self.hdf[-data_len:] = arr

    @property
    def ncol(self) -> int:
        """Number of columns in dataset."""
        return len(self.hdf.dtype)

    @property
    def nrow(self) -> int:
        """Number of rows in dataset."""
        return len(self.hdf)

    @property
    def columns(self) -> Dict[str, Any]:
        """Get dictionary of column specifications."""

        # Get the column definitions from the dataset meta
        metacols = self.meta['columns'] if self.meta is not None and 'columns' in self.meta else {}

        # Get the column names from the dataset
        try:
            dscols = {colname: utils.get_coltype_from_dtype_only(self.hdf.dtype[colname]) for colname in self.hdf.dtype.names}
        except:
            dscols = {}

        # TODO: Handle non-compound datasets

        # Return a merge of the two, with meta cols overridding the dset cols.
        return {**dscols, **metacols}

    @property
    def shape(self) -> Tuple[int, int]:
        """Get dataset shape tuple (rows, cols)."""
        return (self.nrow(), self.ncol())

    def __repr__(self):

        def trunc(string, max_len=20):
            if not isinstance(string, str):
                string = f'{string}'

            if len(string) > max_len:
                return string[:(max_len - 3)] + '...'
            else:
                return string

        lines = []
        ncol = self.ncol
        nrow = self.nrow
        lines.append(f'{self.name}: Dataset [{nrow} rows x {ncol} cols]')
        cols = self.columns
        for col in cols:
            col_meta = cols[col]
            if col_meta['type'] == 'integer':
                tstr = 'integer ({})'.format(
                    'signed' if col_meta['signed'] else 'unsigned')
            elif col_meta['type'] == 'factor':
                nlevels = len(col_meta['levels'])
                lvls = ', '.join([
                    trunc(lvl) for lvl in col_meta['levels'][:min(3, nlevels)]
                ])
                if nlevels > 3:
                    lvls += ', ...'
                tstr = f'factor with {nlevels} levels [{lvls}]'
            else:
                tstr = col_meta['type']
            lines.append(f'  {col}: {tstr}')
        return '\n'.join(lines)

    def __str__(self):
        return self.__repr__()



"""Python API for working with AUViewer."""

import logging
import traceback

from flask import Flask
from pathlib import Path
from typing import List, Dict, Optional

from . import models
from .config import config, set_data_path
from .project import Project
from .shared import createEmptyJSONFile

# Will hold loaded projects
loadedProjects = []

def getProject(id) -> Optional[Project]:
    """
    Returns the project with matching ID.
    :return: the project instance belonging to the id, or None if not found
    """
    global loadedProjects
    for p in loadedProjects:
        if p.id == id:
            return p
    return None

def getProjects() -> Dict[int, Project]:
    """
    Returns loaded projects as a dict.
    :return: dict mapped from project ID to project instance of all loaded projects
    """
    global loadedProjects
    return {p.id: p for p in loadedProjects}

def getProjectsPayload(user_id) -> List[Dict]:
    """
    Returns a list of project information accessible to a given user.
    :return: list of objects containing project information
    """
    global loadedProjects
    return [{
        'id': p.id,
        'name': p.name,
        'files': len(p.files),
        'patterns': p.getTotalPatternCount(),
        'annotations': models.Annotation.query.filter_by(user_id=user_id, project_id=p.id).count(),
        'assignments_rem': models.User.query.filter_by(id=user_id).first().assignments_remaining,
    } for p in loadedProjects]

def listAvailableProjects() -> List[List[str]]:
    """
    Returns list of available projects (ID, name, path) from the database.
    These projects are not necessarily loaded in memory (that is done using
    loadProjects() or loadProject().
    :return:
    """
    return [[p.id, p.name, p.path] for p in models.Project.query.all()]

def listProjects() -> List[List[str]]:
    """
    Returns list of projects loaded in memory (ID, name, path).
    :return: list of lists
    """
    return [[p.id, p.name, p.path] for p in loadedProjects]

def listUsers() -> List[List[str]]:
    """
    Returns list of users (ID, email, first name, last name)
    :return: list of lists
    """
    return [[u.id, u.email, u.first_name, u.last_name] for u in models.User.all()]

def loadProject(id) -> Optional[Project]:
    """
    Load a project into memory. Returns the project instance if successful.
    :return: the project instance, or None
    """
    global loadedProjects

    logging.info(f"Loading project {id}")

    p = models.Project.query.filter_by(id=id).first()

    # Path object for the project
    projDirPathObj = Path(p.path)

    # Validate the project folder
    try:
        validateProjectFolder(projDirPathObj)
    except Exception as e:
        logging.error(f'Project folder {projDirPathObj} is invalid.\n{e}\n{traceback.format_exc()}')
        # TODO(gus): Update the database to reflect this?
    else:
        # TODO(gus): We need to have project take absolute path and project name!
        # Instantiate project, and add to the list to be returned
        loadedProjects.append(Project(p))

    logging.info("Finished loading project.")

    return getProject(id)

def loadProjects() -> Dict[int, Project]:
    """
    Load or reload projects into memory. If new projects are found on disk, they
    will be added to the database & loaded as well. Returns the same output as
    getProjects().
    :return: dict mapped from project ID to project of all loaded projects
    """

    global loadedProjects

    logging.info("Loading projects.")

    # Reset projects to empty list
    loadedProjects = []

    # Load projects from the database
    projs = models.Project.query.all()
    for p in projs:

        logging.info(f"Loading project id {p.id} ({p.name}) at {p.path}")

        # Path object for the project
        projDirPathObj = Path(p.path)

        # Validate the project folder
        try:
            validateProjectFolder(projDirPathObj)
        except Exception as e:
            logging.error(f'Project folder {projDirPathObj} is invalid.\n{e}\n{traceback.format_exc()}')
            # TODO(gus): Update the database to reflect this?
        else:
            # Instantiate project, and add to the list to be returned
            # TODO(gus): We need to have project take absolute path and project name!
            loadedProjects.append(Project(p))

            logging.info("Finished loading project.")

    # Detect new project folders not in the database
    for projDirPathObj in [p for p in config['projectsDirPathObj'].iterdir() if p.is_dir()]:

        # Check whether path exists in database projects
        foundInExistingProject = False
        for p in projs:
            if Path(p.path).samefile(projDirPathObj):
                foundInExistingProject = True
                break

        # Skip if this project folder was found to exist in the database already
        if foundInExistingProject:
            continue

        logging.info(f"Found new project at {projDirPathObj}.")

        # Ensure we have a valid project folder (i.e. it can be empty, but it
        # cannot contain invalid files or folders).
        try:
            validateProjectFolder(projDirPathObj)
        except Exception as e:
            logging.error(f"Project folder {projDirPathObj} is invalid.\n{e}\n{traceback.format_exc()}")
        else:
            logging.info(f'Project folder {projDirPathObj} is valid.')

            # Scaffold project folder
            scaffoldProjectFolder(projDirPathObj)

            # Add project to the database
            project = models.Project(name=projDirPathObj.name, path=str(projDirPathObj.resolve()))
            models.db.session.add(project)
            models.db.session.commit()
            logging.info(f"Added project to database: {projDirPathObj}")

            # Instantiate project
            # TODO(gus): We need to have project take absolute path and project name!
            loadedProjects.append(Project(project))

    logging.info("Finished loading projects.")

    return getProjects()

def scaffoldProjectFolder(projDirPathObj):
    """Generate the baseline project folder contents as needed"""

    logging.info(f"Scaffolding project folder {projDirPathObj}.")

    # Create the originals folder if needed
    (projDirPathObj / 'originals').mkdir(exist_ok=True)

    # Create the processed folder if needed
    (projDirPathObj / 'processed').mkdir(exist_ok=True)

    # Create the templates folder if needed
    projTemplatesFolderPathObj = projDirPathObj / 'templates'
    projTemplatesFolderPathObj.mkdir(exist_ok=True)

    # Create empty template files if needed
    createEmptyJSONFile(projTemplatesFolderPathObj / 'interface_templates.json')
    createEmptyJSONFile(projTemplatesFolderPathObj / 'project_template.json')

    logging.info(f"Finished scaffolding project folder {projDirPathObj}.")

def setDataPath(path, load_projects=False) -> None:
    """
    Loads a data path for use via Python (instead of as a web server). If the
    optional load_projects parameter is False, then projects must be loaded
    manually using loadProject() or loadProjects().
    """

    set_data_path(path)

    app = Flask(__name__)
    app.config.update({
        'SQLALCHEMY_TRACK_MODIFICATIONS': False,
        'SQLALCHEMY_DATABASE_URI': f"sqlite:///{(config['dataPathObj'] / 'database' / 'db.sqlite')}",
    })

    models.init_flask_app(app)
    app.app_context().push()

    if load_projects:
        project.loadProjects()

def validateProjectFolder(projDirPathObj):
    """Raises an exception if the project folder is invalid"""

    logging.info(f"Validating project folder {projDirPathObj}.")

    # Validate top-level folders, excluding dot-files
    for p in [p for p in projDirPathObj.iterdir() if not p.name.startswith('.')]:

        # Project folder should only contain directories
        if not p.is_dir():
            raise Exception(f'Project folder contains invalid file: {p}')

        # Validate expected top-level folder names
        if p.name not in ['originals', 'processed', 'templates']:
            raise Exception(f'Project folder contains invalid folder: {p}')

    # Validate expected template files if they exist.
    tp = p / "templates"
    if tp.exists():
        for p in [p for p in tp.iterdir()]:
            if p.name not in ['project_template.json', 'interface_templates.json']:
                raise Exception(f'Project templates folder contains invalid file: {p}')

    # TODO(gus): Validate that originals have processed?

    logging.info(f'Finished validating project folder {projDirPathObj}.')


"""Class and related functionality for pattern sets."""

from pathlib import Path
from sqlalchemy.orm import joinedload
import pandas as pd
from typing import Union, List

from . import models
from .shared import annotationDataFrame, patternDataFrame

class PatternSet:

    def __init__(self, projparent, dbmodel):

        # Holds reference to the project parent
        self.projparent = projparent

        # Set id & name
        self.id = dbmodel.id
        self.name = dbmodel.name
        self.showByDefault = dbmodel.show_by_default

        # Holds the db model
        self.dbmodel = dbmodel

        # Establish a count of patterns belonging to the set
        self.count = 0
        self.updateCount()

    # Add patterns to the pattern set.
    def addPatterns(self, df):

        # Subset only the columns we need from the user
        df = df[['file_id', 'series', 'left', 'right', 'top', 'bottom', 'label']]

        # Add the project ID
        df['project_id'] = self.projparent.id

        # Add the id of this pattern set
        df.insert(0, "pattern_set_id", [self.id]*df.shape[0])

        # Do the db insert
        df.to_sql('patterns', models.db.engine, index=False, if_exists='append')

        # Update count
        self.updateCount()

    def assignToUsers(self, user_ids: Union[int, List[int]]) -> None:
        """
        Assign the pattern set to user(s).
        :param user_ids: May be single user ID or list of user IDs.
        :return: None
        """
        if not isinstance(user_ids, list):
            user_ids = [user_ids]
        self.dbmodel.users.extend(
            models.User.query.filter(models.User.id.in_(user_ids)).all()
        )
        models.db.session.commit()

    def delete(self, deletePatterns=False):
        """
        Deletes the pattern set from the database and the parent project
        instance. If the pattern set has patterns, the deletion will fail,
        unless the deletePatterns flag is True, in which case it will first
        delete the child patterns.
        """
        if deletePatterns:
            self.deletePatterns()
        models.db.session.rollback()
        try:
            models.db.session.delete(self.dbmodel)
        except:
            models.db.session.rollback()
            raise
        models.db.session.commit()
        del self.projparent.patternsets[self.id]

    def deletePatterns(self) -> int:
        """
        Delete the patterns belonging to this pattern set.
        :return: number of deleted patterns
        """
        models.db.session.rollback()
        try:
            n = models.Pattern.query.filter_by(pattern_set_id=self.id).delete()
        except:
            models.db.session.rollback()
            raise
        models.db.session.commit()
        self.updateCount()
        return n

    def deleteUnannotatedPatterns(self) -> int:
        """
        Delete all patterns which have not yet been annotated from the set.
        :return: number of deleted patterns
        """
        models.db.session.rollback()
        try:
            n = models.Pattern.query.filter(models.Pattern.pattern_set_id == self.id, models.Pattern.id.notin_(
                models.db.session.query(models.Annotation.pattern_id).filter(models.Annotation.pattern_id.isnot(None)).subquery()
            )).delete(synchronize_session=False)
        except:
            models.db.session.rollback()
            raise
        models.db.session.commit()
        self.updateCount()
        return n

    def getAnnotationCount(self) -> int:
        """Returns a count of annotations which annotate any pattern in this set."""
        return models.Annotation.query.filter_by(pattern_set_id=self.id).count()

    def getAnnotations(self) -> pd.DataFrame:
        """Returns a DataFrame of the annotations in this set."""
        return annotationDataFrame(models.Annotation.query.options(joinedload('user')).filter_by(pattern_set_id=self.id).all())

    def getPatternCount(self) -> int:
        """Returns a count of the patterns in this set."""
        return self.count

    def getPatterns(self) -> pd.DataFrame:
        """Returns a DataFrame of the patterns in this set."""
        return patternDataFrame(self.dbmodel.patterns)

    # Set the pattern set's description
    def setDescription(self, description: str):
        """Set the pattern set's description."""
        self.dbmodel.description = description
        models.db.session.commit()

    # Set the pattern set's name
    def setName(self, name: str):
        """Set the pattern set's name."""
        self.dbmodel.name = name
        models.db.session.commit()
        self.name = name

    def setShowByDefault(self, show: bool):
        """Set whether a pattern set should show by default."""
        self.dbmodel.show_by_default = show
        models.db.session.commit()
        self.showByDefault = show

    # Update the count of patterns belonging to this set
    def updateCount(self):
        self.count = models.Pattern.query.filter_by(pattern_set_id=self.id).count()

def getAssignmentsPayload(user_id):
    return [{
        'id': ps.id,
        'name': ps.name,
        'description': ps.description,
        'project_id': ps.project.id,
        'project_name': ps.project.name,
        'completed': models.Annotation.query.filter_by(user_id=user_id, pattern_set_id=ps.id).count(),
        'remaining': len(ps.patterns) - models.Annotation.query.filter_by(user_id=user_id, pattern_set_id=ps.id).count(),
        'total': len(ps.patterns),
    } for ps in models.PatternSet.query.filter(models.PatternSet.users.any(id=user_id)).all()]


from flask import Flask, Blueprint, send_from_directory, request, render_template, render_template_string, abort, Markup
from flask_login import current_user
from flask_mail import Mail
from htmlmin.main import minify
from pathlib import Path
from pprint import pprint
import argparse
import logging
import shutil
import tempfile
import threading
import webbrowser
import json
from flask import jsonify

# Simplejson package is required in order to "ignore" NaN values and implicitly
# convert them into null values. RFC JSON spec left out NaN values, even though
# ES5 supports them (https://www.ecma-international.org/ecma-262/5.1/#sec-4.3.23).
# By default, Python "json" module will allow & json-encode NaN values, but the
# Chrome JS engine will throw an error when trying to parse them. Simplejson
# package, with ignore_nan=True, will implicitly convert NaN values into null
# values. Find "ignore_nan" here: https://simplejson.readthedocs.io/en/latest/
import simplejson

from . import models
from .api import getProject, getProjectsPayload, loadProjects
from .patternset import getAssignmentsPayload
from .config import set_data_path, config, FlaskConfigClass

from .flask_user import confirm_email_required, current_user, login_required, UserManager, SQLAlchemyAdapter
from .flask_user.signals import user_sent_invitation, user_registered

def createApp():

    # Instantiate the Flask web application class
    app = Flask(__name__, template_folder=str(config['codeRootPathObj'] / 'static' / 'www' / 'templates'))

    # Auto-reload templates
    app.jinja_env.auto_reload = True

    # Make the root web path available for templates
    @app.context_processor
    def inject_dict_for_all_templates():
        return {
            'rootWebPath': config['rootWebPath']
        }

    # Minify HTML when possible
    @app.after_request
    def response_minify(response):
        if response.content_type == u'text/html; charset=utf-8':
            response.set_data(minify(response.get_data(as_text=True), remove_comments=True))
        return response

    # Apply Flask configuration for Flask-User package
    app.config.from_object(FlaskConfigClass)
    app.config.update({
        'SECRET_KEY': config['secret_key'],
        'SQLALCHEMY_DATABASE_URI': f"sqlite:///{(config['dataPathObj'] / 'database' / 'db.sqlite')}",
        **config['mail'],
        'USER_CHANGE_PASSWORD_URL': config['rootWebPath'] + app.config['USER_CHANGE_PASSWORD_URL'],
        'USER_CHANGE_USERNAME_URL': config['rootWebPath'] + app.config['USER_CHANGE_USERNAME_URL'],
        'USER_CONFIRM_EMAIL_URL': config['rootWebPath'] + app.config['USER_CONFIRM_EMAIL_URL'],
        'USER_EMAIL_ACTION_URL': config['rootWebPath'] + app.config['USER_EMAIL_ACTION_URL'],
        'USER_FORGOT_PASSWORD_URL': config['rootWebPath'] + app.config['USER_FORGOT_PASSWORD_URL'],
        'USER_INVITE_URL': config['rootWebPath'] + app.config['USER_INVITE_URL'],
        'USER_LOGIN_URL': config['rootWebPath'] + app.config['USER_LOGIN_URL'],
        'USER_LOGOUT_URL': config['rootWebPath'] + app.config['USER_LOGOUT_URL'],
        'USER_MANAGE_EMAILS_URL': config['rootWebPath'] + app.config['USER_MANAGE_EMAILS_URL'],
        'USER_PROFILE_URL': config['rootWebPath'] + app.config['USER_PROFILE_URL'],
        'USER_REGISTER_URL': config['rootWebPath'] + app.config['USER_REGISTER_URL'],
        'USER_RESEND_CONFIRM_EMAIL_URL': config['rootWebPath'] + app.config['USER_RESEND_CONFIRM_EMAIL_URL'],
        'USER_RESET_PASSWORD_URL': config['rootWebPath'] + app.config['USER_RESET_PASSWORD_URL']
    })

    # Initialize the db models with the Flask app
    models.init_flask_app(app)

    # Initialize Flask-Mail
    mail = Mail(app)

    # Setup Flask-User
    db_adapter = SQLAlchemyAdapter(models.db, models.User, UserInvitationClass=models.UserInvitation)  # Select database adapter
    user_manager = UserManager(db_adapter, app)  # Init Flask-User and bind to app

    # You may use this code snippet below to create initial/new admin users
    # (since it is not possible through the interface). Alternatively, you
    # could modify the database.
    #
    with app.app_context():
        new_admin_email = 'gwelter@gmail.com'
        new_admin_pass = 'akeminute'
        if not models.User.query.filter_by(email=new_admin_email).first():
            u = models.User(email=new_admin_email, active=True, password=user_manager.hash_password(new_admin_pass))
            u.roles.append(models.Role(name='admin'))
            models.db.session.add(u)
            models.db.session.commit()

    # Load projects
    with app.app_context():
        loadProjects()

    # Instantiate a file for realtime, in-memory usage (probably temporary)
    # TODO(gus): Refactor realtime
    #rtf = File(projparent=None)

    @user_registered.connect_via(app)
    def after_registered_hook(sender):
        sender.logger.info("USER REGISTERED")

    @user_sent_invitation.connect_via(app)
    def after_invitation_hook(sender):
        sender.logger.info("USER SENT INVITATION")

    ### NON-SECURE AREAS (NO LOGIN REQUIRED) ###

    # Map our static assets to be served
    app.register_blueprint(Blueprint('css', __name__, static_url_path=config['rootWebPath'] + '/css', static_folder=str(config['codeRootPathObj'] / 'static' / 'www' / 'css')))
    app.register_blueprint(Blueprint('fonts', __name__, static_url_path=config['rootWebPath'] + '/fonts', static_folder=str(config['codeRootPathObj'] / 'static' / 'www' / 'fonts')))
    app.register_blueprint(Blueprint('img', __name__, static_url_path=config['rootWebPath'] + '/img', static_folder=str(config['codeRootPathObj'] / 'static' / 'www' / 'img')))
    app.register_blueprint(Blueprint('js', __name__, static_url_path=config['rootWebPath'] + '/js', static_folder=str(config['codeRootPathObj'] / 'static' / 'www' / 'js')))

    @app.route('/favicon.ico')
    def favicon():
        return send_from_directory(str(config['codeRootPathObj'] / 'static' / 'www' / 'img' / 'favicons'), 'favicon.ico')

    ### SECURE AREAS (LOGIN REQUIRED) ###
    ### All methods below should have ###
    ### the @login_required decorator ###

    @app.route(config['rootWebPath']+'/bokeh.html')
    @login_required
    def bokeh():
        return send_from_directory('../www', 'bokeh.html')

    @app.route(config['rootWebPath']+'/create_annotation', methods=['GET'])
    @login_required
    def create_annotation():

        # Parse parameters
        project_id = request.args.get('project_id', type=int)
        file_id = request.args.get('file_id', type=int)
        left = getFloatParamOrNone('xl')
        right = getFloatParamOrNone('xr')
        top = getFloatParamOrNone('yt')
        bottom = getFloatParamOrNone('yb')
        seriesID = request.args.get('sid')
        label = request.args.get('label')
        pattern_id = request.args.get('pattern_id', type=int)

        # Get the project
        project = getProject(project_id)
        if project is None:
            logging.error(f"Project ID {project_id} not found.")
            return app.response_class(
                response=simplejson.dumps({'success': False}),
                status=200,
                mimetype='application/json'
            )

        # Get the file
        file = project.getFile(file_id)
        if file is None:
            logging.error(f"File ID {file_id} not found.")
            return app.response_class(
                response=simplejson.dumps({'success': False}),
                status=200,
                mimetype='application/json'
            )

        # Write the annotation
        newAnnotationID = file.createAnnotation(current_user.id, left, right, top, bottom, seriesID, label, pattern_id)

        # Output response
        return app.response_class(
            response=simplejson.dumps({'success': True, 'id': newAnnotationID}),
            status=200,
            mimetype='application/json'
        )

    @app.route(config['rootWebPath']+'/delete_annotation')
    @login_required
    def delete_annotation():

        # Parse parameters
        id = request.args.get('id', type=int)
        project_id = request.args.get('project_id', type=int)
        file_id = request.args.get('file_id', type=int)

        # Get the project
        project = getProject(project_id)
        if project is None:
            logging.error(f"Project ID {project_id} not found.")
            return app.response_class(
                response=simplejson.dumps({'success': False}),
                status=200,
                mimetype='application/json'
            )

        # Get the file
        file = project.getFile(file_id)
        if file is None:
            logging.error(f"File ID {file_id} not found.")
            return app.response_class(
                response=simplejson.dumps({'success': False}),
                status=200,
                mimetype='application/json'
            )

        # Write the annotation
        deletionSuccess = file.deleteAnnotation(current_user.id, id)

        # Output response
        return app.response_class(
            response=simplejson.dumps({'success': deletionSuccess}),
            status=200,
            mimetype='application/json'
        )

    @app.route(config['rootWebPath']+'/detect_patterns', methods=['GET'])
    @login_required
    def detect_patterns():

        # TODO(gus): Add checks here

        # Parse the series name and alert parameters
        project_id = request.args.get('project_id', type=int)
        file_id = request.args.get('file_id', type=int)
        type = request.args.get('type')
        series = request.args.get('series')
        thresholdlow = request.args.get('thresholdlow', type=float) if request.args.get('thresholdlow') != '' else None
        thresholdhigh = request.args.get('thresholdhigh', type=float) if request.args.get('thresholdhigh') != '' else None
        duration = request.args.get('duration', type=float)
        persistence = request.args.get('persistence', type=float)/100
        maxgap = request.args.get('maxgap', type=float)

        # Get the project
        project = getProject(project_id)
        if project is None:
            logging.error(f"Project ID {project_id} not found.")
            abort(404, description="Project not found.")
            return

        # Get the file
        file = project.getFile(file_id)
        if file is None:
            logging.error(f"File ID {file_id} not found.")
            return app.response_class(
                response=simplejson.dumps([]),
                status=200,
                mimetype='application/json'
            )

        # Run pattern detection
        alerts = file.detectPatterns(
            type=type,
            series=series,
            thresholdlow=thresholdlow,
            thresholdhigh=thresholdhigh,
            duration=duration,
            persistence=persistence,
            maxgap=maxgap
        )

        # Output response
        return app.response_class(
            response=simplejson.dumps(alerts, ignore_nan=True),
            status=200,
            mimetype='application/json'
        )

    @app.route(config['rootWebPath']+'/get_project_annotations')
    @login_required
    def get_project_annotations():

        # Parse parameters
        project_id = request.args.get('project_id', type=int)

        # Get the project
        project = getProject(project_id)
        if project is None:
            logging.error(f"Project ID {project_id} not found.")
            abort(404, description="Project not found.")
            return

        # Assemble project annotations
        projectAnnotations = project.getAnnotationsOutput(current_user.id)

        # Output response
        return app.response_class(
            response=simplejson.dumps(projectAnnotations, ignore_nan=True),
            status=200,
            mimetype='application/json'
        )

    @app.route(config['rootWebPath']+'/')
    @app.route(config['rootWebPath']+'/index.html')
    @login_required
    def index():
        return render_template('index.html', projects=getProjectsPayload(current_user.id), assignments=getAssignmentsPayload(current_user.id))

    @app.route(config['rootWebPath']+'/initial_file_payload')
    @login_required
    def initial_file_payload():

        # Parse parameters
        project_id = request.args.get('project_id', type=int)
        file_id = request.args.get('file_id', type=int)

        # Get the project
        project = getProject(project_id)
        if project is None:
            logging.error(f"Project ID {project_id} not found.")
            abort(404, description="Project not found.")
            return

        # Get the file
        file = project.getFile(file_id)
        if file is None:
            logging.error(f"File ID {file_id} not found.")
            abort(404, description="File not found.")
            return

        # Assemble the initial file payload (full zoomed-out & downsampled, if
        # necessary, datasets for all data series.
        initialFilePayload = file.getInitialPayload(current_user.id)
        
        # Output response
        return app.response_class(
            response=simplejson.dumps(initialFilePayload, ignore_nan=True),
            status=200,
            mimetype='application/json'
        )

    @app.route(config['rootWebPath']+'/project')
    @login_required
    def project():

        # Parse parameters
        id = request.args.get('id', type=int)

        p = getProject(id)
        if p is None:
            logging.error(f"Project ID {id} not found.")
            abort(404, description="Project not found.")
            return

        # Project payload data for the HTML template
        projectPayload = p.getInitialPayload(current_user.id)

        # Assemble the data into a payload. We JSON-encode this twice. The first
        # one converts the dict into JSON. The second one essentially makes the
        # JSON string safe to drop straight into JavaScript code, as we are doing.
        projectPayloadJSON = simplejson.dumps(simplejson.dumps(projectPayload))

        return render_template('project.html', project_name=projectPayload['project_name'], payload=projectPayloadJSON)

    @app.route(config['rootWebPath']+'/series_ranged_data', methods=['GET'])
    @login_required
    def series_ranged_data():

        # Parse parameters
        project_id = request.args.get('project_id', type=int)
        file_id = request.args.get('file_id', type=int)
        series = request.args.getlist('s[]')
        start = request.args.get('start', type=float)
        stop = request.args.get('stop', type=float)

        # Get the project
        project = getProject(project_id)
        if project is None:
            logging.error(f"Project ID {project_id} not found.")
            abort(404, description="Project not found.")
            return

        # Get the file
        file = project.getFile(file_id)
        if file is None:
            logging.error(f"File ID {file_id} not found.")
            abort(404, description="File not found.")
            return

        # Assemble the series ranged data
        seriesRangedData = file.getSeriesRangedOutput(series, start, stop)

        # Output response
        return app.response_class(
            response=simplejson.dumps(seriesRangedData, ignore_nan=True),
            status=200,
            mimetype='application/json'
        )

    @app.route(config['rootWebPath']+'/update_annotation', methods=['GET'])
    @login_required
    def update_annotation():

        # Parse parameters
        project_id = request.args.get('project_id', type=int)
        file_id = request.args.get('file_id', type=int)
        id = request.args.get('id')
        left = getFloatParamOrNone('xl')
        right = getFloatParamOrNone('xr')
        top = getFloatParamOrNone('yt')
        bottom = getFloatParamOrNone('yb')
        seriesID = request.args.get('sid')
        label = request.args.get('label')

        # Get the project
        project = getProject(project_id)
        if project is None:
            logging.error(f"Project ID {project_id} not found.")
            return app.response_class(
                response=simplejson.dumps({ 'success': False }),
                status=200,
                mimetype='application/json'
            )

        # Get the file
        file = project.getFile(file_id)
        if file is None:
            logging.error(f"File ID {file_id} not found.")
            return app.response_class(
                response=simplejson.dumps({ 'success': False }),
                status=200,
                mimetype='application/json'
            )

        # Update the annotation
        updateSuccess = file.updateAnnotation(current_user.id, id, left, right, top, bottom, seriesID, label)

        # Output the response
        return app.response_class(
            response=simplejson.dumps({'success': updateSuccess}),
            status=200,
            mimetype='application/json'
        )

    @app.route(config['rootWebPath'] + '/user_manage')
    @login_required
    def user_manage():
        pprint(vars(current_user))
        return render_template_string("""
                {% extends "base.html" %}
                {% block content %}
                    <h2>Profile Page</h2>
                    <p>Hello {{ current_user.email }},</p>
                    <p><a href="{{ url_for('home_page') }}">Home Page</a></p>
                    <p><a href="{{ url_for('user.change_password') }}">Change password</a></p>
                    <p><a href="{{ url_for('user.invite') }}">Invite User</a></p>
                    <p><a href="{{ url_for('user.logout') }}">Sign out</a></p>
                {% endblock %}
                """)

    ### HELPERS ###

    # Returns the named request parameter as a float or None if the parameter is
    # empty or not present. The default return of None may be overridden.
    def getFloatParamOrNone(name, default=None):
        param = request.args.get(name)
        if param is not None and len(param) > 0:
            return request.args.get(name, type=float)
        else:
            return default

    return app

def main():

    # Parse command-line arguments
    parser = argparse.ArgumentParser(prog='python -m auviewer.serve', description='Auton Lab Universal Viewer')
    parser.add_argument('datapath', type=str, nargs='?', help='Path to data directory (may be empty if starting new)')
    args = parser.parse_args()

    # Handle datapath argument
    if args.datapath is None:
        # If no argument provided, assume ./auvdata
        assumed_datapath = './auvdata'
        logging.warning(f"No datapath argument provided, assuming: {assumed_datapath}")
        set_data_path(assumed_datapath)

    elif Path(args.datapath).is_file():

        # If a file was provided, assume it's one-off view timeseries request
        logging.warning("Running viewer against a single file, using a temp directory for data storage (annotations, etc. will be lost).")

        # Establish paths we'll be working wtih
        temp_datapath = Path(tempfile.gettempdir()) / 'auvdata'
        filedir = temp_datapath / 'projects' / 'temp' / 'originals'
        gtdir = temp_datapath / 'global_templates'
        prjtmplt = gtdir / 'project_template.json'

        target_file = Path(args.datapath).resolve()
        symlink_file = filedir / target_file.name

        logging.info(f"Using temporary data path {temp_datapath}")

        # Delete the previous temp data folder if necessary
        if temp_datapath.is_dir():
            shutil.rmtree(temp_datapath)

        # Create temp data, project, and originals folders recursively
        filedir.mkdir(mode=0o700, parents=True, exist_ok=False)

        # Create global templates folder
        gtdir.mkdir(mode=0o700, exist_ok=False)

        # Create a global project template which shows all series by default
        with prjtmplt.open('x') as f:
            print('{ "series": { "_default": { "show": false } } }', file=f)

        # Establish symlink to the target file
        symlink_file.symlink_to(target_file)

        # Set datapath to our temp datapath
        set_data_path(temp_datapath)

    else:
        # Otherwise, set the data path provided
        set_data_path(args.datapath)

    app = createApp()

    # Open auviewer in the browser, just before we spin up the server
    browser_url = f"http://{config['host']}{':' + str(config['port']) if str(config['port']) != '80' else ''}/{config['rootWebPath']}".rstrip('/')
    if args.datapath is not None and Path(args.datapath).is_file():
        logging.warning("AUViewer is running with a temporary data path! If you create patterns, annotations, or anything else stored in the database, they will be lost!")
        threading.Timer(0.5, lambda: webbrowser.open(browser_url+'/project?id=1#file_id=1')).start()
    else:
        threading.Timer(0.5, lambda: webbrowser.open(browser_url)).start()

    app.run(host=config['host'], port=config['port'], debug=config['debug'], use_reloader=False)


# Start development web server
if __name__ == '__main__':
    main()
